import re
from http import HTTPStatus
from typing import Any
from typing import Literal
from typing import TypedDict

import aiohttp
from aiohttp import ClientError
from aiohttp import ClientResponse
from aiohttp import ClientSession

from netzooe_eservice_api.constants import COMMON_HEADERS
from netzooe_eservice_api.constants import ConsentsStatus
from netzooe_eservice_api.constants import ConsumptionsProfilesBranch
from netzooe_eservice_api.constants import ESERVICE_PORTAL
from netzooe_eservice_api.constants import ESERVICE_PORTAL_API
from netzooe_eservice_api.error import APIError
from netzooe_eservice_api.error import AuthenticationError


class Pod(TypedDict):
    contract_account_number: str
    energy_community_id: str
    profile_type: str
    best_available_granularity: str
    meter_point_administration_number: str
    date_from: str
    date_to: str


class NetzOOEeServiceAPI:
    """An asynchronous client to interact with the Netz OÖ eService API."""

    def __init__(
        self,
        username: str,
        password: str,
        *,
        session: aiohttp.ClientSession,
    ) -> None:
        """Initialize API with username and password.

        Parameters
        ----------
        username
            eService username
        password
            eService password
        session
            Add an aiohttp client session

        Examples
        --------
        >>> async with ClientSession() as session:
        >>>     client = NetzOOEeServiceAPI(
        >>>         username="test",
        >>>         password="test",
        >>>         session=session,
        >>>     )

        """
        self._username = username
        self._password = password
        self._session: ClientSession = session
        self.xsrf_token: str = ""

    @property
    def headers(self) -> dict[str, str]:
        """Default headers all API calls."""
        return {
            **COMMON_HEADERS,
            "Content-Type": "application/json",
            "x-xsrf-token": self.xsrf_token,
        }

    async def _get_session(self) -> ClientResponse:
        async with self._session.get(f"{ESERVICE_PORTAL_API}/session", headers={**COMMON_HEADERS}) as resp:
            if resp.status != HTTPStatus.OK:
                await self._session.close()
                raise APIError(status=HTTPStatus(resp.status))

            return resp

    def _set_xsrf_token(self, headers) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN001
        cookies: list[str] = headers.getall("Set-Cookie", [])
        cookie_string: str = ";".join(cookies)

        if match := re.search(r"XSRF-TOKEN=([^;]+)", cookie_string):
            self.xsrf_token = match.group(1)

    async def _request(
        self,
        method: Literal["GET", "POST"],
        url: str,
        *,
        json: Any | None = None,  # noqa: ANN401
        retry: bool = True,
    ) -> Any:  # noqa: ANN401
        if not self.xsrf_token:
            await self.login()

        headers: dict[str, str] = self.headers.copy()

        if json is not None:
            headers["Content-Type"] = "application/json"

        try:
            async with self._session.request(method, url, headers=headers, json=json) as resp:
                if resp.status == HTTPStatus.OK:
                    return await resp.json()
                if resp.status == HTTPStatus.UNAUTHORIZED:
                    if retry:
                        await self.login()
                        return await self._request(method, url, json=json, retry=False)

                    await self._session.close()
                    raise AuthenticationError(status=HTTPStatus.UNAUTHORIZED)

                message = await resp.text()
                raise APIError(message, status=HTTPStatus(resp.status))
        except ClientError as error:
            await self._session.close()
            raise APIError(str(error)) from error

    async def _get(self, url: str) -> Any:  # noqa: ANN401
        return await self._request("GET", url)

    async def _post(self, url: str, /, *, json: Any) -> Any:  # noqa: ANN401
        return await self._request("POST", url, json=json)

    async def login(self) -> None:
        """Authenticate to the Netz OÖ eService portal."""
        async with await self._session.post(
            f"{ESERVICE_PORTAL}/service/j_security_check",
            json={
                "j_username": self._username,
                "j_password": self._password,
            },
            headers={
                **COMMON_HEADERS,
                "Content-Type": "application/json",
            },
        ) as resp:
            if resp.status != HTTPStatus.OK:
                await self._session.close()

                if resp.status == HTTPStatus.UNAUTHORIZED:
                    raise AuthenticationError(status=HTTPStatus.UNAUTHORIZED)

                raise APIError(status=HTTPStatus(resp.status))

            _session: ClientResponse = await self._get_session()
            self._set_xsrf_token(_session.headers)

    async def logout(self) -> None:
        """Logout from the Netz OÖ eService portal."""
        await self._session.get(
            f"{ESERVICE_PORTAL}/service/logout",
            headers={
                **COMMON_HEADERS,
            },
        )

        if not self._session.closed:
            await self._session.close()

    async def dashboard(self) -> dict[str, Any]:
        """Get data from the eService dashboard."""
        data: dict[str, Any] = await self._get(f"{ESERVICE_PORTAL_API}/dashboard")
        return data

    async def consents(self, status: list[ConsentsStatus] | ConsentsStatus | None = None, /) -> list[dict[str, Any]]:
        """Get data from the eService data sharing."""
        _status: str = ""

        if status is not None and not isinstance(status, list):
            status = [status]

        if isinstance(status, list):
            _status = ",".join([_.value for _ in status])
            _status = f"?status={_status}"

        data: list[dict[str, Any]] = await self._get(f"{ESERVICE_PORTAL_API}/consents{_status}")
        return data

    async def consumptions_profiles(
        self, branch: list[ConsumptionsProfilesBranch] | ConsumptionsProfilesBranch | None
    ) -> list[dict[str, Any]]:
        """Get data from the eService profiles."""
        _branch: str = ""

        if branch is not None and not isinstance(branch, list):
            branch = [branch]

        if isinstance(branch, list):
            _branch = ",".join([_.value for _ in branch])
            _branch = f"?branch={_branch}"

        data: list[dict[str, Any]] = await self._get(f"{ESERVICE_PORTAL_API}/consumptions/profiles{_branch}")
        return data

    async def contract_accounts(self, business_partner_number: str, contract_account_number: str) -> dict[str, Any]:
        """Get data from the eService contract accounts."""
        data: dict[str, Any] = await self._get(
            f"{ESERVICE_PORTAL_API}/contract-accounts/{business_partner_number}/{contract_account_number}"
        )
        return data

    async def consumptions_profile(self, pods: list[Pod]) -> list[dict[str, Any]]:
        """Get data from the eService consumptions profile."""
        data: list[dict[str, Any]] = await self._post(
            f"{ESERVICE_PORTAL_API}/consumptions/profile/active",
            json={
                "pods": [
                    {
                        "contractAccountNumber": pod["contract_account_number"],
                        "energyCommunityId": pod["energy_community_id"],
                        "type": pod["profile_type"],
                        "bestAvailableGranularity": pod["best_available_granularity"],
                        "meterPointAdministrationNumber": pod["meter_point_administration_number"],
                        "timerange": {
                            "from": pod["date_from"],
                            "to": pod["date_to"],
                        },
                    }
                    for pod in pods
                ],
                "dimension": "ENERGY",
            },
        )
        return data



from typing import Any
from typing import Final

import pytest
from aiohttp import ClientError
from aiohttp import ClientSession
from aioresponses import aioresponses
from yarl import URL

from netzooe_eservice_api.api import NetzOOEeServiceAPI
from netzooe_eservice_api.api import Pod
from netzooe_eservice_api.constants import ConsentsStatus
from netzooe_eservice_api.constants import ConsumptionsProfilesBranch
from netzooe_eservice_api.error import APIError
from netzooe_eservice_api.error import AuthenticationError

LOGIN_AND_LOGOUT_API_CALLS: Final[int] = 3


async def create_client(mock_api: aioresponses, /, *, repeat: bool | int = False) -> NetzOOEeServiceAPI:
    session: ClientSession = ClientSession()

    mock_api.post(
        "https://eservice.netzooe.at/service/j_security_check",
        status=200,
        repeat=repeat,
    )

    mock_api.get(
        "https://eservice.netzooe.at/service/v1.0/session",
        headers={
            "Set-Cookie": "XSRF-TOKEN=mocked-token-value; Path=/; Secure",
        },
        status=200,
        repeat=repeat,
    )

    mock_api.get(
        "https://eservice.netzooe.at/service/logout",
        status=200,
        repeat=False,
    )

    client: NetzOOEeServiceAPI = NetzOOEeServiceAPI(
        username="test",
        password="test",  # noqa: S106
        session=session,
    )

    return client


async def create_logged_in_client(mock_api: aioresponses, /, *, repeat: bool | int = False) -> NetzOOEeServiceAPI:
    client: NetzOOEeServiceAPI = await create_client(mock_api, repeat=repeat)

    await client.login()
    return client


class TestHappyPathNetzOOEeServiceAPI:
    @pytest.mark.asyncio
    async def test_login(self) -> None:
        with aioresponses() as mock_api:
            client: NetzOOEeServiceAPI = await create_logged_in_client(mock_api)
            await client.logout()

            assert client.xsrf_token == "mocked-token-value"  # noqa: S105

            login_key = ("POST", URL("https://eservice.netzooe.at/service/j_security_check"))
            session_key = ("GET", URL("https://eservice.netzooe.at/service/v1.0/session"))

            assert login_key in mock_api.requests
            assert session_key in mock_api.requests

            assert len(mock_api.requests) == LOGIN_AND_LOGOUT_API_CALLS
            assert len(mock_api.requests[login_key]) == 1
            assert len(mock_api.requests[session_key]) == 1

            login_call = mock_api.requests[login_key][0]

            assert login_call.kwargs["json"] == {
                "j_username": "test",
                "j_password": "test",
            }

            assert login_call.kwargs["headers"] == {
                "User-Agent": "Mozilla/5.0",
                "Accept": "application/json, text/plain, */*",
                "client-id": "netzonline",
                "Content-Type": "application/json",
            }

            session_call = mock_api.requests[session_key][0]

            assert session_call.kwargs["headers"] == {
                "User-Agent": "Mozilla/5.0",
                "Accept": "application/json, text/plain, */*",
                "client-id": "netzonline",
            }

    @pytest.mark.asyncio
    async def test_reconnect(self) -> None:
        with aioresponses() as mock_api:
            client: NetzOOEeServiceAPI = await create_logged_in_client(mock_api, repeat=2)

            mock_api.get("https://eservice.netzooe.at/service/v1.0/dashboard", status=401, payload={})
            mock_api.get("https://eservice.netzooe.at/service/v1.0/dashboard", status=200, payload={})

            result = await client.dashboard()
            assert result == {}

            await client.logout()

            dashboard_key = ("GET", URL("https://eservice.netzooe.at/service/v1.0/dashboard"))
            assert dashboard_key in mock_api.requests

            assert len(mock_api.requests) == LOGIN_AND_LOGOUT_API_CALLS + 1
            assert len(mock_api.requests[dashboard_key]) == 2  # noqa: PLR2004

    @pytest.mark.asyncio
    async def test_dashboard(self) -> None:
        with aioresponses() as mock_api:
            client: NetzOOEeServiceAPI = await create_logged_in_client(mock_api)

            mock_api.get("https://eservice.netzooe.at/service/v1.0/dashboard", status=200, payload={})

            result = await client.dashboard()
            assert result == {}

            await client.logout()

            dashboard_key = ("GET", URL("https://eservice.netzooe.at/service/v1.0/dashboard"))
            assert dashboard_key in mock_api.requests

            assert len(mock_api.requests) == LOGIN_AND_LOGOUT_API_CALLS + 1
            assert len(mock_api.requests[dashboard_key]) == 1

    @pytest.mark.asyncio
    async def test_dashboard_without_login(self) -> None:
        with aioresponses() as mock_api:
            client: NetzOOEeServiceAPI = await create_client(mock_api)

            mock_api.get("https://eservice.netzooe.at/service/v1.0/dashboard", status=200, payload={})

            result = await client.dashboard()
            assert result == {}

            await client.logout()

            dashboard_key = ("GET", URL("https://eservice.netzooe.at/service/v1.0/dashboard"))
            assert dashboard_key in mock_api.requests

            assert len(mock_api.requests) == LOGIN_AND_LOGOUT_API_CALLS + 1
            assert len(mock_api.requests[dashboard_key]) == 1

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ("status", "url"),
        [
            (
                [
                    ConsentsStatus.ACTIVE,
                    ConsentsStatus.ACTIVE_UNCHANGEABLE,
                ],
                "https://eservice.netzooe.at/service/v1.0/consents?status=ACTIVE%252CACTIVE_UNCHANGEABLE",
            ),
            (
                ConsentsStatus.ACTIVE,
                "https://eservice.netzooe.at/service/v1.0/consents?status=ACTIVE",
            ),
        ],
    )
    async def test_consents(self, status: list[ConsentsStatus] | ConsentsStatus | None, url: str) -> None:
        with aioresponses() as mock_api:
            client: NetzOOEeServiceAPI = await create_logged_in_client(mock_api)

            mock_api.get(url, status=200, payload=[])

            result: list[dict[str, Any]] = await client.consents(status)
            assert result == []

            await client.logout()

            consents_key = ("GET", URL(url))
            assert consents_key in mock_api.requests

            assert len(mock_api.requests) == LOGIN_AND_LOGOUT_API_CALLS + 1
            assert len(mock_api.requests[consents_key]) == 1

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ("branch", "url"),
        [
            (
                None,
                "https://eservice.netzooe.at/service/v1.0/consumptions/profiles",
            ),
            (
                ConsumptionsProfilesBranch.ELECTRICITY,
                "https://eservice.netzooe.at/service/v1.0/consumptions/profiles?branch=STROM",
            ),
        ],
    )
    async def test_consumptions_profiles(
        self, branch: list[ConsumptionsProfilesBranch] | ConsumptionsProfilesBranch | None, url: str
    ) -> None:
        with aioresponses() as mock_api:
            client: NetzOOEeServiceAPI = await create_logged_in_client(mock_api)

            mock_api.get(url, status=200, payload=[])

            result: list[dict[str, Any]] = await client.consumptions_profiles(branch)
            assert result == []

            await client.logout()

            consumptions_profiles_key = ("GET", URL(url))
            assert consumptions_profiles_key in mock_api.requests

            assert len(mock_api.requests) == LOGIN_AND_LOGOUT_API_CALLS + 1
            assert len(mock_api.requests[consumptions_profiles_key]) == 1

    @pytest.mark.asyncio
    async def test_contract_account(self) -> None:
        with aioresponses() as mock_api:
            client: NetzOOEeServiceAPI = await create_logged_in_client(mock_api)

            mock_api.get("https://eservice.netzooe.at/service/v1.0/contract-accounts/123/345", status=200, payload={})

            result = await client.contract_accounts(business_partner_number="123", contract_account_number="345")
            assert result == {}

            await client.logout()

            contract_account_key = ("GET", URL("https://eservice.netzooe.at/service/v1.0/contract-accounts/123/345"))
            assert contract_account_key in mock_api.requests

            assert len(mock_api.requests) == LOGIN_AND_LOGOUT_API_CALLS + 1
            assert len(mock_api.requests[contract_account_key]) == 1

    @pytest.mark.asyncio
    async def test_consumptions_profile(self) -> None:
        with aioresponses() as mock_api:
            client: NetzOOEeServiceAPI = await create_logged_in_client(mock_api)

            mock_api.post(
                "https://eservice.netzooe.at/service/v1.0/consumptions/profile/active", status=200, payload=[]
            )

            result: list[dict[str, Any]] = await client.consumptions_profile(
                pods=[
                    Pod(
                        contract_account_number="123",
                        energy_community_id="345",
                        profile_type="mocked-profile",
                        best_available_granularity="mocked-granularity",
                        meter_point_administration_number="AT123",
                        date_from="2026-01-01",
                        date_to="2026-01-30",
                    )
                ]
            )
            assert result == []

            await client.logout()

            consumptions_profile_key = (
                "POST",
                URL("https://eservice.netzooe.at/service/v1.0/consumptions/profile/active"),
            )
            assert consumptions_profile_key in mock_api.requests

            assert len(mock_api.requests) == LOGIN_AND_LOGOUT_API_CALLS + 1
            assert len(mock_api.requests[consumptions_profile_key]) == 1

            consumptions_profile_call = mock_api.requests[consumptions_profile_key][0]

            payload = consumptions_profile_call.kwargs["json"]

            assert payload["dimension"] == "ENERGY"

            pod: dict[str, Any] = payload["pods"][0]
            assert pod["contractAccountNumber"] == "123"
            assert pod["energyCommunityId"] == "345"
            assert pod["type"] == "mocked-profile"
            assert pod["bestAvailableGranularity"] == "mocked-granularity"
            assert pod["meterPointAdministrationNumber"] == "AT123"
            assert pod["timerange"]["from"] == "2026-01-01"
            assert pod["timerange"]["to"] == "2026-01-30"


class TestUnhappyPathNetzOOEeServiceAPI:

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ("exception", "status", "message"),
        [
            (
                AuthenticationError,
                401,
                "401 Unauthorized: No permission -- see authorization schemes",
            ),
            (
                APIError,
                400,
                "400 Bad Request: Bad request syntax or unsupported method",
            ),
        ],
    )
    async def test_login_failed(self, exception: type[APIError], status: int, message: str) -> None:
        with aioresponses() as mock_api:
            mock_api.post(
                "https://eservice.netzooe.at/service/j_security_check",
                status=status,
            )
            session: ClientSession = ClientSession()

            client: NetzOOEeServiceAPI = NetzOOEeServiceAPI(
                username="test",
                password="test",  # noqa: S106
                session=session,
            )

            with pytest.raises(exception) as error:
                await client.login()

            assert str(error.value) == message
            assert client.xsrf_token == ""

            login_key = ("POST", URL("https://eservice.netzooe.at/service/j_security_check"))

            assert len(mock_api.requests) == 1
            assert login_key in mock_api.requests
            assert len(mock_api.requests[login_key]) == 1

            login_call = mock_api.requests[login_key][0]

            assert login_call.kwargs["json"] == {
                "j_username": "test",
                "j_password": "test",
            }

            assert login_call.kwargs["headers"] == {
                "User-Agent": "Mozilla/5.0",
                "Accept": "application/json, text/plain, */*",
                "client-id": "netzonline",
                "Content-Type": "application/json",
            }

    @pytest.mark.asyncio
    async def test_session_failed(self) -> None:
        with aioresponses() as mock_api:
            mock_api.post(
                "https://eservice.netzooe.at/service/j_security_check",
                status=200,
            )

            mock_api.get(
                "https://eservice.netzooe.at/service/v1.0/session",
                status=401,
            )

            session: ClientSession = ClientSession()

            client: NetzOOEeServiceAPI = NetzOOEeServiceAPI(
                username="test",
                password="test",  # noqa: S106
                session=session,
            )

            with pytest.raises(APIError) as error:
                await client.login()

            assert str(error.value) == "401 Unauthorized: No permission -- see authorization schemes"

            assert client.xsrf_token == ""

            login_key = ("POST", URL("https://eservice.netzooe.at/service/j_security_check"))
            session_key = ("GET", URL("https://eservice.netzooe.at/service/v1.0/session"))

            assert login_key in mock_api.requests
            assert session_key in mock_api.requests

            assert len(mock_api.requests) == 2  # noqa: PLR2004
            assert len(mock_api.requests[login_key]) == 1
            assert len(mock_api.requests[session_key]) == 1

    @pytest.mark.asyncio
    async def test_reconnect_failed(self) -> None:
        with aioresponses() as mock_api:
            client: NetzOOEeServiceAPI = await create_logged_in_client(mock_api, repeat=2)

            mock_api.get("https://eservice.netzooe.at/service/v1.0/dashboard", status=401, repeat=2, payload={})

            with pytest.raises(APIError) as error:
                await client.dashboard()

            assert str(error.value) == "401 Unauthorized: No permission -- see authorization schemes"

    @pytest.mark.asyncio
    async def test_client_error(self) -> None:
        with aioresponses() as mock_api:
            client: NetzOOEeServiceAPI = await create_logged_in_client(mock_api)

            mock_api.get(
                "https://eservice.netzooe.at/service/v1.0/dashboard",
                exception=ClientError("mocked error"),
            )

            with pytest.raises(APIError) as error:
                await client.dashboard()

            assert str(error.value) == "mocked error"

    @pytest.mark.asyncio
    async def test_internal_server_error(self) -> None:
        with aioresponses() as mock_api:
            client: NetzOOEeServiceAPI = await create_logged_in_client(mock_api)

            mock_api.get(
                "https://eservice.netzooe.at/service/v1.0/dashboard",
                status=500,
            )

            with pytest.raises(APIError) as error:
                await client.dashboard()

            assert str(error.value) == "500 Internal Server Error: Server got itself in trouble"



from enum import Enum

ESERVICE_PORTAL: str = "https://eservice.netzooe.at"
ESERVICE_PORTAL_API: str = f"{ESERVICE_PORTAL}/service/v1.0"

COMMON_HEADERS: dict[str, str] = {
    "User-Agent": "Mozilla/5.0",
    "Accept": "application/json, text/plain, */*",
    "client-id": "netzonline",
}


class ConsentsStatus(Enum):
    ACTIVE = "ACTIVE"
    ACTIVE_UNCHANGEABLE = "ACTIVE_UNCHANGEABLE"
    REVOKED = "REVOKED"


class ConsumptionsProfilesBranch(Enum):
    ELECTRICITY = "STROM"


class SynthProfile(Enum):
    HOUSEHOLD = "Haushalt"
    PHOTOVOLTAICS = "Photovoltaik"
